/**
 * Undo stack and Command base.
 */
(function() {

  // based on Backbone.js' inherits 
  var ctor = function() {};
  var inherits = function(parent, protoProps) {
    var child;

    if (protoProps && protoProps.hasOwnProperty('constructor')) {
      child = protoProps.constructor;
    } else {
      child = function() {
        return parent.apply(this, arguments);
      };
    }

    ctor.prototype = parent.prototype;
    child.prototype = new ctor();

    if (protoProps) extend(child.prototype, protoProps);

    child.prototype.constructor = child;
    child.__super__ = parent.prototype;
    return child;
  };

  function extend(target, ref) {
    var name, value;
    for (name in ref) {
      value = ref[name];
      if (value !== undefined) {
        target[name] = value;
      }
    }
    return target;
  };

  var Undo;
  if (typeof exports !== 'undefined') {
    Undo = exports;
  } else {
    Undo = this.Undo = {};
  }

  Undo.Stack = function() {
    this.commands = [];
    this.stackPosition = -1;
    this.savePosition = -1;
  };

  extend(Undo.Stack.prototype, {
    execute: function(command) {
      this._clearRedo();
      var ret = command.execute();
      this.commands.push(command);
      this.stackPosition++;
      this.changed();
      console.log('execute: ' + command.name + ' ' + ret);
      return ret;
    },
    undo: function() {
      var command = this.commands[this.stackPosition];
      command.undo();
      this.stackPosition--;
      this.changed();
      console.log('undo: ' + command.name);
    },
    canUndo: function() {
      return this.stackPosition >= 0;
    },
    redo: function() {
      this.stackPosition++;
      var command = this.commands[this.stackPosition];
      command.redo();
      this.changed();
      console.log('redo: ' + command.name);
    },
    canRedo: function() {
      return this.stackPosition < this.commands.length - 1;
    },
    save: function() {
      this.savePosition = this.stackPosition;
      this.changed();
    },
    dirty: function() {
      return this.stackPosition != this.savePosition;
    },
    _clearRedo: function() {
      // TODO there's probably a more efficient way for this
      this.commands = this.commands.slice(0, this.stackPosition + 1);
    },
    changed: function() {
      // do nothing, override
    },
    getTopCommand: function() {
      if (this.canUndo())
        return this.commands[this.stackPosition];
      else
        return null;
    }
  });

  Undo.Command = function(name) {
    this.name = name;
  }

  var up = new Error("override me!");

  extend(Undo.Command.prototype, {
    execute: function() {
      throw up;
    },
    undo: function() {
      throw up;
    },
    redo: function() {
      this.execute();
    },
    stickUniqueName: function(target) {
      if (target.length == 0) return '';

      var targetName = target.attr(this.uniqueNameAttr);
      if (targetName == undefined || targetName == '') {
        targetName = 'undo' + FoxUtils.getUniqueName();
        target.attr(this.uniqueNameAttr, targetName); 
      }
      
      return "[" + this.uniqueNameAttr + "='" + targetName + "']";
    }
  });

  Undo.Command.extend = function(protoProps) {
    var child = inherits(this, protoProps);
    child.extend = Undo.Command.extend;
    return child;
  };

}).call(this);