//
// Custom command.
//
var CustomCommand = Undo.Command.extend({
	name: 'CustomCommand',
	constructor: function() {},
	
	execute: function() {
		console.log('execute: ' + this.name);
	},

	undo: function() {
		console.log('undo: ' + this.name);
	},

	redo: function() {
		this.execute();	
	}
});

//
// Resize image
//
var ResizeImageCommand = Undo.Command.extend({
	name: 'ResizeImageCommand',
	constructor: function(image, newWidth, newHeight) {
		this.target = image;
		this.oldWidth = image.width();
		this.oldHeight = image.height();
		this.newWidth = newWidth;
		this.newHeight = newHeight;
	},
		
	execute: function() {
		this.newWidth ? this.target.attr('width', this.newWidth) : this.target.attr('width', 'auto');
		this.newHeight ? this.target.attr('height', this.newHeight) : this.target.attr('height', 'auto');
		return true;
	},

	undo: function() {
		this.target.attr('width', this.oldWidth);
		this.target.attr('height', this.oldHeight);
	},

	redo: function() {
		this.execute();	
	}
});

/////////////////////////// Table ////////////////////////////////

//
// Table command.
//
var TableCommand = Undo.Command.extend({
	name: 'TableCommand',
	uniqueNameAttr: 'uniqueName',

	getTargetCell: function() {
		var selection =  window.getSelection();
		var cell = $(selection.anchorNode).closest('td');
		if (cell.length == 0) {
			cell = $(selection.anchorNode).closest('th');
		}
		return cell;
	},

	selectClass: 'selected-cell',
	getSelectCells: function() {
		return $('.' + TableSelect.selectClass);	
	}
});

//
// Delete table.
//
var DeleteTableCommand = TableCommand.extend({
	name: 'DeleteTableCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;
		var tableContainer = targetCell.closest('.table-container');
		if (tableContainer.length == 0) return false;
		
		var focusNode = $(window.getSelection().focusNode);
		while (focusNode.length !== 0 && focusNode[0].nodeType == 3) { // None text node
			focusNode = focusNode.parent();
		}
		focusNode.addClass('lastFocusNode');

		var that = this;

		var prevName = this.stickUniqueName(tableContainer.prev());
		var parentName = this.stickUniqueName(tableContainer.parent());

		this.removeNote = {
			tableContainer: tableContainer.clone(),
			prevName: prevName,
			parentName: parentName
		}
		
		tableContainer.remove();

		return true;
	},

	undo: function() {
		if (this.targetName == '') return;

		if (this.removeNote.prevName !== '') {
			$(this.removeNote.prevName).after(this.removeNote.tableContainer);
		} else {
			$(this.removeNote.parentName).prepend(this.removeNote.tableContainer);
		}

		var lastFocusNode = this.removeNote.tableContainer.find('.lastFocusNode');
		if (lastFocusNode.length !== 0) {
			setTimeout(function() {
				window.getSelection().collapse(lastFocusNode[0], 0);
				lastFocusNode.removeClass('lastFocusNode');
			}, 10);
		}

		TableInit.initTable(this.removeNote.tableContainer.find('table'));
	},

	redo: function() {
		this.execute();
	}
});

//
// Insert table caption.
//
var InsertTableCaptionCommand = TableCommand.extend({
	name: 'InsertTableCaptionCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		targetCell.closest('table').prepend('<caption style="border: 1px dashed #DDD; border-bottom: 0; padding: 3px;">&#8203;</caption>');
		return true;
	},

	undo: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;
		targetCell.closest('table').find('caption').remove();
	},

	redo: function() {
		this.execute();
	}
});

//
// Delete table caption.
//
var DeleteTableCaptionCommand = TableCommand.extend({
	name: 'DeleteTableCaptionCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell(), 'cell');
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		var caption = targetCell.closest('table').find('caption');
		this.removeCaption = caption.clone();
		caption.remove();
		return true;
	},

	undo: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		targetCell.closest('table').prepend(this.removeCaption);
	},

	redo: function() {
		this.execute();
	}
});

//
// Insert header row.
//
var InsertHeaderRowCommand = TableCommand.extend({
	name: 'InsertHeaderRowCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');
		var length = tableCore.matrix[0].length;

		var border = targetCell.css('border');
		var padding = targetCell.css('padding');
		var header = $('<tr style="background-color: rgb(210, 234, 241);"></tr>');
		for (var i = 0; i < length; i++) {
			header.append('<th style="border: ' + border + '; padding: ' + padding + ';" rowspan="1" colspan="1">&#8203;</th>');
		}

		targetCell.closest('table').prepend(header);

		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		var tr = targetCell.closest('table').find('tr').first();
		if (tr.find('td').length == 0) {	// is header row?
			tr.remove();
		}

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Delete header row.
//
var DeleteHeaderRowCommand = TableCommand.extend({
	name: 'DeleteHeaderRowCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		var tr = targetCell.closest('table').find('tr').first();
		if (tr.find('td').length == 0) {	// is header row?
			this.removeHeaderRow = tr.clone();
			tr.remove();
		}

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		targetCell.closest('table').prepend(this.removeHeaderRow);

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Insert header column.
//
var InsertHeaderColumnCommand = TableCommand.extend({
	name: 'InsertHeaderColumnCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		var border = targetCell.css('border');
		var padding = targetCell.css('padding');
		targetCell.closest('table').find('tr').each(function() {
			$(this).prepend('<th style="background-color: rgb(210, 234, 241); width: 50px; border: ' + border + '; padding: '+ padding + ';" rowspan="1" colspan="1">&#8203;</th>');
		});

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		targetCell.closest('table').find('tr').each(function() {
			$(this).find('th').first().remove();
		});

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Delete header column.
//
var DeleteHeaderColumnCommand = TableCommand.extend({
	name: 'DeleteHeaderColumnCommand',
	constructor: function() {
		this.targetName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		this.removeNotes = [];
		var that = this;
		targetCell.closest('table').find('tr').each(function() {
			var tr = $(this);
			var trName = that.stickUniqueName(tr);

			var th = tr.find('th').first();
			that.removeNotes.push({
				target: th.clone(),
				trName: tr
			});
			th.remove();
		});

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetName);
		if (targetCell.length == 0) return false;

		for (var i = 0; i < this.removeNotes.length; i++) {
			var note = this.removeNotes[i];
			$(note.trName).prepend(note.target);
		}

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();	
	},

	redo: function() {
		this.execute();
	}
});

//
// Delete current column.
//
var DeleteCurrentColumnCommand = TableCommand.extend({
	name: 'DeleteCurrentColumnCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
		var tableContainer = this.getTargetCell().closest('.table-container');
		this.targetContainerName = this.stickUniqueName(tableContainer);
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;
		var targetContainer = $(this.targetContainerName);
		if (targetContainer.length == 0) return false;

		var tableCore = targetContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row,
			column: coord.column
		};

		this.removeNotes = [];
		this.changeNotes = [];
		var that = this;

		tableCore.getThinColumnByCoord(atomCoord).each(function() {
			var cell = $(this);
			var currentCoord = cell.data('coord');
			
			if (currentCoord.colspan == 1) {
				var prevName = that.stickUniqueName(cell.prev());
				var parentName = that.stickUniqueName(cell.parent());

				that.removeNotes.push({
					target: cell.clone(),
					prevName: prevName,
					parentName: parentName
				});
				cell.remove();
			} else {
				var targetName = that.stickUniqueName(cell);

				that.changeNotes.push({
					targetName: targetName,
					colspan: currentCoord.colspan
				});
				cell.attr('colspan', currentCoord.colspan - 1);
			}
		});

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetContainer = $(this.targetContainerName);
		if (targetContainer.length == 0) return false;

		var tableCore = targetContainer.data('TableCore');

		for (var i = 0; i < this.removeNotes.length; i++) {
			var note = this.removeNotes[i];
			if (note.prevName !== '') {
				$(note.prevName).after(note.target);
			} else {
				$(note.parentName).prepend(note.target);	
			}
		}

		for (var i = 0; i < this.changeNotes.length; i++) {
			var note = this.changeNotes[i];
			$(note.targetName).attr('colspan', note.colspan);
		}

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();	
	}
});

//
// Delete current row.
//
var DeleteCurrentRowCommand = TableCommand.extend({
	name: 'DeleteCurrentRowCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');
		
		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row,
			column: coord.column
		};

		var isDeleteRow = false;
		this.changeNodes = [];
		var that = this;

		tableCore.getThinRowByCoord(atomCoord).each(function() {
			var cell = $(this);
			var currentCoord = cell.data('coord');

			if (currentCoord.rowspan === 1) {
				isDeleteRow = true;
			} else {
				var targetName = this.stickUniqueName(cell);
				that.changeNodes.push({
					targetName: targetName,
					rowspan: currentCoord.rowspan
				});
				cell.attr('rowspan', currentCoord.rowspan - 1);
			}
		});

		if (isDeleteRow) {
			var tr = targetCell.closest('tr');
			var prevName = this.stickUniqueName(tr.prev());
			var parentName = this.stickUniqueName(tr.parent());

			this.removeNote = {
				target: tr.clone(),
				prevName: prevName,
				parentName: parentName
			};

			tr.remove();
		}

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		if (this.removeNote) {
			if (this.removeNote.prevName !== '') {
				$(this.removeNote.prevName).after(this.removeNote.target);
			} else {
				$(this.removeNote.parentName).prepend(this.removeNote.target);
			}
		}

		for (var i = 0; i < this.changeNodes.length; i++) {
			var note = this.changeNodes[i];
			$(note.targetName).attr('rowspan', note.rowspan);
		}

		// 需要重置单元格矩阵
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;
		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Insert column in the left side.
//
var InsertLeftColumnCommand = TableCommand.extend({
	name: 'InsertLeftColumnCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row,
			column: coord.column
		};

		var that = this;

		// 找到当前列，在当前列的td前面插入新的td。
		tableCore.getThinColumnByCoord(atomCoord).each(function() {
			var cell = $(this);
			var currentCoord = cell.data('coord');
			var currentAtomCoord = {
				row: currentCoord.row,
				column: currentCoord.column
			};

			if (atomCoord.column === currentAtomCoord.column) {
				var newCell = cell.clone().attr('colspan', 1).attr(that.uniqueNameAttr, '').html('&#8203;');
				cell.before(newCell);
			} else {
				cell.attr('colspan', currentCoord.colspan + 1);
			}
		});

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row,
			column: coord.column
		};

		// 找到当前列，删除当前列左侧的td。
		tableCore.getThinColumnByCoord(atomCoord).each(function() {
			var td = $(this);
			var currentCoord = td.data('coord');
			var currentAtomCoord = {
				row: currentCoord.row,
				column: currentCoord.column
			};

			if (atomCoord.column === currentAtomCoord.column) {
				td.prev().remove();
			} else {
				td.attr('colspan', currentCoord.colspan - 1);
			}
		});

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();	
	}
});

//
// Insert column in the right side.
//
var InsertRightColumnCommand = TableCommand.extend({
	name: 'InsertRightColumnCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row,
			column: coord.column + coord.colspan - 1
		};

		var that = this;

		tableCore.getThinColumnByCoord(atomCoord).each(function() {
			var cell = $(this);
			var currentCoord = cell.data('coord');
			var currentAtomCoord = {
				row: currentCoord.row,
				column: currentCoord.column + currentCoord.colspan - 1
			};

			if (atomCoord.column === currentAtomCoord.column) {
				var newCell = cell.clone().attr('colspan', 1).attr(that.uniqueNameAttr, '').html('&#8203;');
				cell.after(newCell);
			} else {
				cell.attr('colspan', currentCoord.colspan + 1);
			}
		});

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row,
			column: coord.column + coord.colspan - 1
		};

		tableCore.getThinColumnByCoord(atomCoord).each(function() {
			var td = $(this);
			var currentCoord = td.data('coord');
			var currentAtomCoord = {
				row: currentCoord.row,
				column: currentCoord.column + currentCoord.colspan - 1
			};

			if (atomCoord.column === currentAtomCoord.column) {
				td.next().remove();
			} else {
				td.attr('colspan', currentCoord.colspan - 1);
			}
		});

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();	
	}
});

//
// Insert row in the top side.
//
var InsertTopRowCommand = TableCommand.extend({
	name: 'InsertTopRowCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');
		var coord = targetCell.data('coord');
		
		var atomCoord = {
			row: coord.row,
			column: coord.column
		};

		this.changeNodes = [];
		var that = this;

		var newRow = $('<tr></tr>');
		tableCore.getThinRowByCoord(atomCoord).each(function() {
			var cell = $(this);
			var currentCoord = cell.data('coord');
			
			var currentCell = {
				row: currentCoord.row,
				column: currentCoord.column
			};

			if (atomCoord.row === currentCell.row) {
				var newCell = cell.clone().html('&#8203;').attr('rowspan', 1).attr(that.uniqueNameAttr, '');
				newRow.append(newCell);
			} else {
				var targetName = this.stickUniqueName(cell);
				this.changeNodes.push({
					targetName: targetName,
					rowspan: currentCell.rowspan
				});
				cell.attr('rowspan', currentCoord.rowspan + 1);
			}
		});

		targetCell.closest('tr').before(newRow);

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		for (var i = 0; i < this.changeNodes.length; i++) {
			var note = this.changeNodes[i];
			$(note.targetName).attr('rowspan', note.rowspan);
		}
		targetCell.closest('tr').prev().remove();

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Insert row in the bottom side.
//
var InsertBottomRowCommand = TableCommand.extend({
	name: 'InsertBottomRowCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var atomCoord = {
			row: coord.row + coord.rowspan - 1,
			column: coord.column
		};

		this.changeNodes = [];
		var that = this;

		var newRow = $('<tr></tr>');

		tableCore.getThinRowByCoord(atomCoord).each(function() {
			var cell = $(this);
			var currentCoord = cell.data('coord');
			var currentAtomCoord = {
				row: currentCoord.row + currentCoord.rowspan - 1,
				column: currentCoord.column
			};

			if (atomCoord.row === currentAtomCoord.row) {
				var newTd = cell.clone().html('&#8203;').attr('rowspan', 1).attr(that.uniqueNameAttr, '');
				newRow.append(newTd);
			} else {
				var targetName = this.stickUniqueName(cell);
				this.changeNodes.push({
					targetName: targetName,
					rowspan: currentAtomCoord.rowspan
				});
				cell.attr('rowspan', currentCoord.rowspan + 1);
			}
		});

		var tr = targetCell.closest('tr');
		for (var i = 1; i < coord.rowspan; i++) {
			tr = tr.next();	
		}
		tr.after(newRow);

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var tableContainer = targetCell.closest('.table-container');
		var tableCore = tableContainer.data('TableCore');

		var coord = targetCell.data('coord');
		var tr = targetCell.closest('tr');
		for (var i = 1; i < coord.rowspan; i++) {
			tr = tr.next();	
		}
		tr.next().remove();

		for (var i = 0; i < this.changeNodes.length; i++) {
			var note = this.changeNodes[i];
			$(note.targetName).attr('rowspan', note.rowspan);
		}

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Merge cells.
//
var MergeCellCommand = TableCommand.extend({
	name: 'MergeCellCommand',
	constructor: function() {
		this.targetCellNames = [];
		var that = this;

		this.getSelectCells().each(function() {
			that.targetCellNames[that.targetCellNames.length] = that.stickUniqueName($(this));
		});	
	},

	getTargetCells: function() {
		var cells = $('div').filter('td');
		for (var i = 0; i < this.targetCellNames.length; i++) {
			cells = $.merge(cells, $(this.targetCellNames[i]));
		}
		return cells;
	},

	execute: function() {
		var targetCells = this.getTargetCells();
		if (targetCells.length == 0) return false;
		targetCells.removeClass(this.selectClass);

		var surround = {
			minRow:	9999,
			minColumn: 9999,
			maxRow: 0,
			maxColumn: 0
		}
		
		targetCells.each(function() {
			var cell = $(this);
			var coord = cell.data('coord');
			if (surround.minRow > coord.row) surround.minRow = coord.row;
			if (surround.minColumn > coord.column) surround.minColumn = coord.column;
			if (surround.maxRow < coord.row + coord.rowspan - 1) surround.maxRow = coord.row + coord.rowspan - 1;
			if (surround.maxColumn < coord.column + coord.colspan - 1) surround.maxColumn = coord.column + coord.colspan - 1;
		});
		

		var table = targetCells.closest('.table-container');
		var tableCore = table.data('TableCore');
		var changeCell = tableCore.matrix[surround.minRow][surround.minColumn];
		var coord = changeCell.data('coord');

		this.removeNotes = [];
		var that = this;

		targetCells.each(function() {
			var cell = $(this);
			if (cell[0] !== changeCell[0]) {
				var prevName = that.stickUniqueName(cell.prev());
				var parentName = that.stickUniqueName(cell.parent());

				that.removeNotes.push({
					target: cell.clone(),
					prevName: prevName,
					parentName: parentName,
				});
				cell.remove();
			}
		});

		var targetName = this.stickUniqueName(changeCell);
		this.changeNote = {
			targetName: targetName,
			rowspan: coord.rowspan,
			colspan: coord.colspan
		};

		changeCell.attr('rowspan', surround.maxRow - surround.minRow + 1);
		changeCell.attr('colspan', surround.maxColumn - surround.minColumn + 1);
		window.getSelection().collapse(changeCell[0], 0);

		// 需要重置单元格矩阵
		tableCore.setTableMatrix();

		return true;
	},

	undo: function() {
		var targetCells = this.getTargetCells();
		if (targetCells.length == 0) return false;

		for (var i = this.removeNotes.length - 1; i >= 0; i--) {
			var note = this.removeNotes[i];
			if (note.prevName !== '') {
				$(note.prevName).after(note.target);
			} else {
				$(note.parentName).prepend(note.target);
			}
		}

		var note = this.changeNote;
		$(note.targetName).attr('rowspan', note.rowspan).attr('colspan', note.colspan);

		// 需要重置单元格矩阵
		var tableCore = targetCells.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Split cells.
//
var SplitCellCommand = TableCommand.extend({
	name: 'SplitCellCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		var coord = targetCell.data('coord');
		if (coord.rowspan == 1 && coord.colspan == 1) return false;

		var tableCore = targetCell.closest('.table-container').data('TableCore');
		var coord = targetCell.data('coord');

		var targetName = this.stickUniqueName(targetCell);
		this.changeNote = {
			targetName: targetName,
			rowspan: coord.rowspan,
			colspan: coord.colspan
		};

		var that = this;
		this.addNotes = [];

		// 分裂列
		var columnCells = tableCore.NoneElement();
		columnCells = $.merge(columnCells, targetCell);

		for (var i = 1; i < coord.colspan; i++) {
			var newCell = targetCell.clone().html('').attr(this.uniqueNameAttr, '');
			newCell.attr('colspan', 1);
			targetCell.after(newCell);
			columnCells = $.merge(columnCells, newCell);

			var targetName = this.stickUniqueName(newCell);
			this.addNotes.push({
				targetName: targetName
			});
		}

		targetCell.attr('colspan', 1);
		tableCore.setTableMatrix();

		// 分裂行
		var trs = targetCell.closest('table').find('tr');
		columnCells.each(function() {
			var cell = $(this);
			var coord = cell.data('coord');
			cell.attr('rowspan', 1);
			
			for (var row = coord.row + 1; row < coord.row + coord.rowspan; row++) {
				var column = coord.column - 1;
				var beforeCell = null;
				var hasBeforeCell = false;
				while (column >= 0) {
					beforeCell = tableCore.matrix[row][column];
					var beforeCoord = beforeCell.data('coord');
					if (beforeCoord.rowspan == 1) {
						hasBeforeCell = true;
						break;
					}
					column--;
				}

				var newCell = cell.clone().html('').attr('rowspan', 1).attr(that.uniqueNameAttr, '');
				if (hasBeforeCell) {
					beforeCell.after(newCell);	
				} else {
					$(trs[row]).prepend(newCell);
				}
				var targetName = that.stickUniqueName(newCell);
				that.addNotes.push({
					targetName: targetName
				});
			}

			tableCore.setTableMatrix();
		});

		return true;
	},

	undo: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;

		for (var i = 0; i < this.addNotes.length; i++) {
			var note = this.addNotes[i];
			$(note.targetName).remove();
		}
		
		var note = this.changeNote;
		$(note.targetName).attr('rowspan', note.rowspan).attr('colspan', note.colspan);

		// 需要重置单元格矩阵
		var tableCore = targetCell.closest('.table-container').data('TableCore');
		tableCore.setTableMatrix();
	},

	redo: function() {
		this.execute();
	}
});

//
// Interlace color for rows.
//
var InterlaceRowColorCommand = TableCommand.extend({
	name: 'InterlaceRowColorCommand',
	constructor: function() {
		this.targetCellName = this.stickUniqueName(this.getTargetCell());
	},

	execute: function() {
		var targetCell = $(this.targetCellName);
		if (targetCell.length == 0) return false;
		
		var color = [
			'rgb(252, 252, 252)',
			//'rgb(247, 250, 255)'
			'rgb(178, 221, 243)'
		];

		this.changeNodes = [];
		var that = this;

		var colorIndex = 0;
		targetCell.closest('table').find('tr').each(function() {
			var tr = $(this);
			if (tr.find('td').length > 0) {
				var targetName = that.stickUniqueName(tr);
				that.changeNodes.push({
					targetName: targetName,
					color: tr.css('background-color')
				});
				tr.css('background-color', color[colorIndex % 2]);
				colorIndex++;
			}
		});

		return true;
	},

	undo: function() {
		for (var i = 0; i < this.changeNodes.length; i++) {
			var note = this.changeNodes[i];
			$(note.targetName).css('background-color', note.color);
		}	
	},

	redo: function() {
		this.execute();
	}
});

